#ifndef STRINGSOURCE_H
#define STRINGSOURCE_H

#include <frost_utils.h>
#include <frost_utils_file.h>

namespace Frost
{

class InputSourceBase
{
public :

    InputSourceBase();
    virtual ~InputSourceBase();

    virtual s_bool IsValid() const = 0;
    virtual char Get() = 0;
};

template<class T>
class Source : public InputSourceBase
{
};

template<>
class Source<File> : public InputSourceBase
{
public :

    Source(File& mFile);

    s_bool IsValid() const;
    char Get();

private :

    File& mFile_;
};

template<>
class Source<s_str> : public InputSourceBase
{
public :

    Source(const s_str& sString);

    s_bool IsValid() const;
    char Get();

private :

    const s_str& sString_;
    s_uint       uiPos_;
};

class InputSource
{
public :

    template<class T>
    InputSource(T& mImpl) : mImpl_(new Source<T>(mImpl))
    {
    }

    template<class T>
    InputSource(const T& mImpl) : mImpl_(new Source<T>(mImpl))
    {
    }

    s_bool IsValid() const;
    char Get();

    void Ignore(const char& c);

private :

    s_ctnr<char> lIgnoredList_;
    s_refptr<InputSourceBase> mImpl_;
};

}

#endif
