#ifndef MODULE_H
#define MODULE_H

#include <frost_utils.h>

#include "class.h"
#include "context.h"

namespace Frost
{

class Compiler;

class Module
{
public :

    Module(s_ptr<Compiler> pCompiler, const s_str& sFile);

    void SetSelf(s_wptr<Module> pSelf);

    void SetDocElements(const s_ctnr<s_str>& lDocElements);

    void SetName(const TypeName& mName);
    void SetPreloaded();
    void SetHeaderFile(const s_str& sHeader);

    void Compile();

    const s_bool& IsPreloaded();
    const s_str& GetName() const;
    const s_str& GetFullName() const;
    s_str GetCreatedSourceFile() const;
    s_str GetCreatedHeaderFile() const;

    void UsesModule(s_wptr<Module> pModule);
    void UsesNamespace(const s_ctnr<s_str>& lNamespace);
    void UsesType(const s_ctnr<s_str>& lType);
    void UsesContent(const s_str& sContent);

    void Exports(const s_str& sContent);

    void SetScope(const s_str& sScope);

    void CreateMainClass();
    s_wptr<Class> BeginClass(const s_str& sName, const s_str& sClassType);
    void EndClass();
    s_wptr<Class> GetCurrentClass();

    void AddVariable(Variable mVariable);
    void AddFunction(Function mFunction);
    void AddTypedef(Typedef mTypedef);
    void AddEnum(Enum mEnum);
    void AddInheritance(TypeName mClass);

    void DecorateType(TypeName& mType, const s_ctnr<s_str>& lContext) const;

    void Serialize() const;

    static s_str MakeFileName(const TypeName& mName);

    void CompileScope_(Scope& s, const s_ctnr<s_str>& lContext);

protected :

    s_str GetForwardDeclarations() const;
    s_str GetHeaderFile() const;
    s_str GetUsingDirectives() const;

    void CheckType_(const TypeName& sTypeName);
    s_ctnr<TypeName> DecorateType_(const TypeName& mType, const s_ctnr<s_str>& lContext) const;
    void HeaderNeedsModule_(s_wptr<Module> pModule);

    s_str Beautify_(const s_str& sStr, const s_uint& uiIndent = 0) const;
    void  SerializeClassHeader_(File& mHeader, const Class& c) const;
    void  SerializeClassSource_(File& mSource, const Class& c) const;
    s_str SerializeFunctionHeader_(const Function& f, const s_bool& bTemplate = false) const;
    s_str SerializeFunctionParamsAndBody_(const Function& f) const;
    s_str SerializeDocumentation_(const s_ctnr<s_str>& lDocElements) const;
    s_str SerializeTypedef_(const Typedef& t) const;

    s_bool IsReachable_(const TypeName& mType) const;

    s_wptr<Module> pSelf_;

    s_bool bPreloaded_;
    s_str  sFile_;

    s_refptr<Class> pMainClass_;
    TypeName        mName_;
    s_str           sHeaderFile_;

    s_ctnr<s_wptr<Class>> lClassStack_;
    s_wptr<Class>         pCurrentClass_;
    Scope                 mGlobalScope_;

    s_map<s_str, Context> lContextList_;
    s_ptr<Context>        pCurrentContext_;

    s_ctnr<TypeName>      lUsedTypeList_;
    s_ctnr<s_ctnr<s_str>> lUsedNamespaceList_;

    s_ctnr<s_wptr<Module>> lUsedModuleList_;
    s_ctnr<s_str>          lExternalModuleList_;

    s_ctnr<s_str> lExportedContent_;
    s_ctnr<s_str> lInternalContent_;

    s_ptr<Compiler> pCompiler_;

};

}

#endif // MODULE_H
