#include "class.h"
#include "module.h"
#include "compiler.h"
#include "function.h"
#include "variable.h"
#include "typedef.h"
#include "enum.h"

using namespace Frost;

Class::Class(s_ptr<Compiler> pCompiler, s_wptr<Module> pModule, const s_str& sType) :
    sClassType_(sType), pCompiler_(pCompiler), pModule_(pModule)
{
    Scope s;

    if (sType == "class")
        s.sName = "private";
    else
        s.sName = "public";

    s.pClassParent = this;

    // TODO : ## REWRITE EVERYTHING !! ##
    // ----------------------------------
    // 1) A class HAS scopes or a class IS a scope ?
    //    Rewrite Class accordingly
    // 2) Rewrite Module with scopes
    //    Add support for namespaces
    //    Rewrite DecorateType (rename to FindType ?)
    //    Implement UsesNamespace
    //    Implement UsesType

    lScopeList_.PushBack(s);
    pCurrentScope_ = &lScopeList_.Back();
}

const s_ctnr<Scope>& Class::GetScopeList() const
{
    return lScopeList_;
}

void Class::SetDocElements(const s_ctnr<s_str>& lDocElements)
{
    lDocElements_ = lDocElements;
}

const s_ctnr<s_str>& Class::GetDocElements() const
{
    return lDocElements_;
}

void Class::SetType(const TypeName& mName)
{
    mType_ = mName;
}

const TypeName& Class::GetType() const
{
    return mType_;
}

const s_str& Class::GetClassType() const
{
    return sClassType_;
}

const TemplateParams& Class::GetTemplateParams() const
{
    return mTemplateParams_;
}

void Class::SetScope(const s_str& sScope)
{
    if (pCurrentScope_->lVariableList_.IsEmpty() &&
        pCurrentScope_->lFunctionList_.IsEmpty() &&
        pCurrentScope_->lClassList_.IsEmpty())
    {
        lScopeList_.PopBack();
    }

    Scope s;
    s.sName_ = sScope;
    lScopeList_.PushBack(s);
    pCurrentScope_ = &lScopeList_.Back();
}

s_wptr<Class> Class::AddClass(const s_str& sName, const s_str& sType)
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, iterScope->lClassList_)
        {
            if ((*iterClass)->GetType().sType == sName)
                pCompiler_->ThrowError("a class with the name \'" + sName + "\' already exists");
        }
    }

    pCurrentScope_->lClassList_.PushBack(s_refptr<Class>(new Class(pCompiler_, pModule_, sType)));
    s_wptr<Class> pClass = pCurrentScope_->lClassList_.Back();

    TypeName mName;
    mName.sType = sName;
    mName.sFullType = mType_.sFullType + "::" + sName;
    mName.sNakedType = sName;
    mName.sFullNakedType = mName.sFullType;
    mName.sNamespace = mType_.sNamespace;
    mName.lNamespaces = mType_.lNamespaces;
    mName.sMembership = (mType_.sMembership.IsEmpty() ? mType_.sType : mType_.sMembership + "::" + mType_.sType);
    mName.lMemberships = mType_.lMemberships;
    mName.lMemberships.PushBack(mType_.sType);
    mName.sNestedType = mName.sMembership + "::" + mName.sType;
    mName.bIsNested = true;
    mName.pOwner = this;
    mName.pClass = pClass;

    if (pCurrentScope_->sName_ == "private")
        mName.bIsPrivate = true;
    else if (pCurrentScope_->sName_ == "protected")
        mName.bIsProtected = true;

    pClass->SetType(mName);

    return pClass;
}

void Class::AddVariable(const Variable& mVariable)
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<Variable>::const_iterator iterVar;
        foreach (iterVar, iterScope->lVariableList_)
        {
            if (iterVar->sName == mVariable.sName)
                pCompiler_->ThrowError("a variable with the name \'" + mVariable.sName + "\' already exists");
        }
    }

    pCurrentScope_->lVariableList_.PushBack(mVariable);
}

void Class::AddFunction(const Function& mFunction)
{
    if (mFunction.bDestructor)
    {
        s_ctnr<Scope>::const_iterator iterScope;
        foreach (iterScope, lScopeList_)
        {
            s_ctnr<Function>::const_iterator iterFunc;
            foreach (iterFunc, iterScope->lFunctionList_)
            {
                if (iterFunc->bDestructor)
                    pCompiler_->ThrowError("a destructor already exists");
            }
        }
    }

    pCurrentScope_->lFunctionList_.PushBack(mFunction);
}

s_ptr<Typedef> Class::AddTypedef(const Typedef& mTypedef)
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<Typedef>::const_iterator iterType;
        foreach (iterType, iterScope->lTypedefList_)
        {
            if (iterType->mType.sType == mTypedef.mType.sType)
                pCompiler_->ThrowError("a typedef with the name \'" + mTypedef.mType.sType + "\' already exists");
        }
    }

    pCurrentScope_->lTypedefList_.PushBack(mTypedef);

    TypeName& mName = pCurrentScope_->lTypedefList_.Back().mType;
    mName.sFullType = mType_.sFullType + "::" + mName.sType;
    mName.sNakedType = mName.sType;
    mName.sFullNakedType = mName.sFullType;
    mName.sNamespace = mType_.sNamespace;
    mName.lNamespaces = mType_.lNamespaces;
    mName.sMembership = (mType_.sMembership.IsEmpty() ? mType_.sType : mType_.sMembership + "::" + mType_.sType);
    mName.lMemberships = mType_.lMemberships;
    mName.lMemberships.PushBack(mType_.sType);
    mName.sNestedType = mName.sMembership + "::" + mName.sType;
    mName.bIsNested = true;
    mName.pOwner = this;
    mName.bIsTypedef = true;

    if (pCurrentScope_->sName_ == "private")
        mName.bIsPrivate = true;
    else if (pCurrentScope_->sName_ == "protected")
        mName.bIsProtected = true;

    return &pCurrentScope_->lTypedefList_.Back();
}

void Class::AddEnum(const Enum& mEnum)
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<Enum>::const_iterator iterEnum;
        foreach (iterEnum, iterScope->lEnumList_)
        {
            if (iterEnum->mType.sType == mEnum.mType.sType)
                pCompiler_->ThrowError("an enum with the name \'" + mEnum.mType.sType + "\' already exists");
        }
    }

    pCurrentScope_->lEnumList_.PushBack(mEnum);

    TypeName& mName = pCurrentScope_->lEnumList_.Back().mType;
    mName.sFullType = mType_.sFullType + "::" + mName.sType;
    mName.sNakedType = mName.sType;
    mName.sFullNakedType = mName.sFullType;
    mName.sNamespace = mType_.sNamespace;
    mName.lNamespaces = mType_.lNamespaces;
    mName.sMembership = (mType_.sMembership.IsEmpty() ? mType_.sType : mType_.sMembership + "::" + mType_.sType);
    mName.lMemberships = mType_.lMemberships;
    mName.lMemberships.PushBack(mType_.sType);
    mName.sNestedType = mName.sMembership + "::" + mName.sType;
    mName.bIsNested = true;
    mName.pOwner = this;
    mName.bIsEnum = true;

    if (pCurrentScope_->sName_ == "private")
        mName.bIsPrivate = true;
    else if (pCurrentScope_->sName_ == "protected")
        mName.bIsProtected = true;
}

void Class::AddInheritance(const TypeName& mClass)
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<TypeName>::const_iterator iterInh;
        foreach (iterInh, iterScope->lInheritanceList_)
        {
            if (mClass.sFullType == iterInh->sFullType)
                pCompiler_->ThrowError("already inherits from '"+mClass.sFullType+"' (in "+iterScope->sName_+" scope)");
        }
    }

    pCurrentScope_->lInheritanceList_.PushBack(mClass);
}

void Class::SetTemplateParams(const TemplateParams& mParams)
{
    bIsTemplate_ = true;
    mTemplateParams_ = mParams;
}

s_ctnr<TypeName> Class::DecorateType(const TypeName& mType, const s_ctnr<s_str>& lContext) const
{
    s_ctnr<TypeName> lList;
    TypeName mTemp;

    if (mType_.TryDecorateType(mType, mTemp, pModule_, lContext))
        lList.PushBack(mTemp);

    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<Typedef>::const_iterator iterType;
        foreach (iterType, iterScope->lTypedefList_)
        {
            if (iterType->mType.TryDecorateType(mType, mTemp, pModule_, lContext))
                lList.PushBack(mTemp);
        }
        s_ctnr<Enum>::const_iterator iterEnum;
        foreach (iterEnum, iterScope->lEnumList_)
        {
            if (iterEnum->mType.TryDecorateType(mType, mTemp, pModule_, lContext))
                lList.PushBack(mTemp);
        }
    }

    foreach (iterScope, lScopeList_)
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, iterScope->lClassList_)
        {
            s_ctnr<TypeName> lTempList = (*iterClass)->DecorateType(mType, lContext);
            if (!lTempList.IsEmpty())
                lList.PushBack(lTempList);
        }
    }

    return lList;
}

s_ctnr<s_ptr<const Class>> Class::GetClassList() const
{
    s_ctnr<s_ptr<const Class>> lClassList;
    lClassList.PushBack(this);

    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<s_refptr<Class>>::const_iterator iterClass;
        foreach (iterClass, iterScope->lClassList_)
        {
            lClassList.PushBack((*iterClass)->GetClassList());
        }
    }

    return lClassList;
}

const s_bool& Class::IsTemplate() const
{
    return bIsTemplate_;
}

void Class::CompileScopes(s_ctnr<s_str>& lContext)
{
    lContext.PushBack(mType_.sType);
    s_ctnr<Scope>::iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        pModule_->CompileScope_(*iterScope, lContext);

        s_ctnr<s_refptr<Class>>::iterator iterClass;
        foreach (iterClass, iterScope->lClassList_)
            (*iterClass)->CompileScopes(lContext);
    }
    lContext.PopBack();
}

s_bool Class::InheritsFrom(s_wptr<Class> pClass) const
{
    s_ctnr<Scope>::const_iterator iterScope;
    foreach (iterScope, lScopeList_)
    {
        s_ctnr<TypeName>::const_iterator iterInh;
        foreach (iterInh, iterScope->lInheritanceList_)
        {
            if (iterInh->pClass == pClass || iterInh->pClass->InheritsFrom(pClass))
                return true;
        }
    }

    return false;
}
